import cgi
import json
import tempfile

import pytest

from bddrest import Given, when, story, response, InvalidUrlParametersError, \
    CallVerifyError, Call, AlteredCall, Story


def wsgi_application(environ, start_response):
    form = cgi.FieldStorage(
        fp=environ['wsgi.input'],
        environ=environ,
        strict_parsing=False,
        keep_blank_values=True
    )

    try:
        # FIXME: Why x ^ 1234
        code = int(form['activationCode'].value) ^ 1234
    except ValueError:
        start_response('400 Bad Request', [('Content-Type', 'text/plain;utf-8')])
        return

    start_response('200 OK', [
        ('Content-Type', 'application/json;charset=utf-8'),
        ('X-Pagination-Count', '10')
    ])
    result = dict(
        secret='ABCDEF',
        code=code,
        query=environ['QUERY_STRING'],
    )

    identity = environ.get('HTTP_AUTHORIZATION')
    if identity:
        result['identity'] = identity
    yield json.dumps(result).encode()


def test_given_when():
    call = dict(
        title='Binding and registering the device after verifying the activation code',
        description=\
            'As a new visitor I have to bind my device with activation code and phone number',
        url='/apiv1/devices/name: SM-12345678',
        verb='POST',
        as_='visitor',
        query=dict(
            a=1,
            b=2
        ),
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        )
    )
    with Given(wsgi_application, **call) as s:
        assert s.response is response._resolver()
        assert response.status == '200 OK'
        assert response.status == 200
        assert 'secret' in response.json
        assert response.json['secret'] == 'ABCDEF'
        assert 'Bad Header' not in response.headers

        assert response.content_type == 'application/json'
        assert response.json ==  dict(
            code=745525,
            secret='ABCDEF',
            query='a=1&b=2'
        )

        when(
            'Trying invalid code',
            form=dict(
                activationCode='badCode'
            )
        )

        assert response.status == 400


def test_url_parameters():
    call = dict(
        title='Multiple url parameters',
        url='/apiv1/devices/name: SM-12345678/id: 1',
        verb='POST',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
    )

    with Given(wsgi_application, **call):
        assert response.status == '200 OK'

        with pytest.raises(InvalidUrlParametersError):
            when(
                title='Incomplete url parameters',
                url_parameters=dict(
                    id=2
                )
            )

        with pytest.raises(InvalidUrlParametersError):
            when(
                title='Extra url parameters',
                url_parameters=dict(
                    name='any',
                    id=3,
                    garbage='yes'
                )
            )

        with pytest.raises(InvalidUrlParametersError):
            when(
                title='Without url parameters',
                url_parameters=None
            )

    call = dict(
        title='No url parameters',
        url='/apiv1/devices',
        verb='POST',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
    )

    with Given(wsgi_application, **call):
        assert response.status == '200 OK'

def test_to_dict():
    call = dict(
        title='Binding',
        description='Awesome given description',
        url='/apiv1/devices/name: SM-12345678',
        verb='POST',
        as_='visitor',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
        headers=[('X-H1', 'Header Value')]
    )
    with Given(wsgi_application, **call):
        assert response.status == '200 OK'
        when(
            'Trying invalid code',
            description='Awesome invalid code description',
            form=dict(
                activationCode='badCode'
            )
        )
        assert response.status == 400

        story_dict = story.to_dict()
        assert story_dict['base_call'] == dict(
            title='Binding',
            description='Awesome given description',
            url='/apiv1/devices/:name',
            verb='POST',
            as_='visitor',
            url_parameters=dict(name='SM-12345678'),
            form=dict(
                activationCode='746727',
                phone='+9897654321'
            ),
            headers=['X-H1: Header Value'],
            response=dict(
                status='200 OK',
                headers=[
                    'Content-Type: application/json;charset=utf-8',
                    'X-Pagination-Count: 10'
                ],
                json={'secret': 'ABCDEF', 'code': 745525, 'query': ''}
            )
        )
        assert story_dict['calls'][0] == dict(
            title='Trying invalid code',
            description='Awesome invalid code description',
            form=dict(
                activationCode='badCode'
            ),
            response=dict(
                headers=['Content-Type: text/plain;utf-8'],
                status='400 Bad Request',
            )
        )


def test_from_dict():
    data = dict(
        base_call=dict(
            title='Binding',
            url='/apiv1/devices/:name',
            verb='POST',
            as_='visitor',
            url_parameters=dict(name='SM-12345678'),
            form=dict(
                activationCode='746727',
                phone='+9897654321'
            ),
            headers=['X-H1: Header Value'],
            response=dict(
                status='200 OK',
                headers=[
                    'Content-Type: application/json;charset=utf-8',
                    'X-Pagination-Count: 10'
                ],
                json={'secret': 'ABCDEF', 'code': 745525, 'query': ''}
            )
        ),
        calls=[
            dict(
                title='Trying invalid code',
                form=dict(
                    activationCode='badCode'
                ),
                response=dict(
                    headers=['Content-Type: text/plain;utf-8'],
                    status='400 Bad Request',
                )
            )
        ]
    )
    loaded_story = Story.from_dict(data)
    assert loaded_story is not None
    assert isinstance(loaded_story.base_call, Call)
    assert isinstance(loaded_story.calls[0], AlteredCall)
    assert loaded_story.base_call.response.status == 200
    assert data == loaded_story.to_dict()


def test_dump_load():
    call = dict(
        title='Binding',
        url='/apiv1/devices/name: SM-12345678',
        verb='POST',
        as_='visitor',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
        headers=[('X-H1', 'Header Value')]
    )
    with Given(wsgi_application, **call):
        assert response.status == '200 OK'
        when(
            'Trying invalid code',
            form=dict(
                activationCode='badCode'
            )
        )
        assert response.status == 400

        dumped_story = story.dumps()
        loaded_story = Story.loads(dumped_story)
        assert story.to_dict() == loaded_story.to_dict()
        loaded_story.validate()


def test_verify():
    call = dict(
        title='Binding',
        url='/apiv1/devices/name: SM-12345678',
        verb='POST',
        as_='visitor',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
        headers=[('X-H1', 'Header Value')]
    )
    with Given(wsgi_application, **call):
        assert response.status == '200 OK'
        when(
            'Trying invalid code',
            form=dict(
                activationCode='badCode'
            )
        )
        dumped_story = story.dumps()

    loaded_story = Story.loads(dumped_story)
    loaded_story.verify(wsgi_application)

    loaded_story.base_call.response.body = '{"a": 1}'

    with pytest.raises(CallVerifyError):
        loaded_story.verify(wsgi_application)


def test_dump_load_file():
    with tempfile.TemporaryFile(mode='w+', encoding='utf-8') as temp_file:
        call = dict(
            title='Binding',
            url='/apiv1/devices/name: SM-12345678',
            verb='POST',
            as_='visitor',
            form=dict(
                activationCode='746727',
                phone='+9897654321'
            ),
            headers=[('X-H1', 'Header Value')]
        )
        with Given(wsgi_application, **call):
            assert response.status == '200 OK'
            when(
                'Trying invalid code',
                form=dict(
                    activationCode='badCode'
                )
            )
            story.dump(temp_file)

        temp_file.seek(0)
        loaded_story = Story.load(temp_file)
        assert loaded_story.verify(wsgi_application) is None


def test_url_overriding():
    call = dict(
        title='Multiple url parameters',
        url='/apiv1/devices/name: SM-12345678/id: 1',
        verb='POST',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
    )

    with Given(wsgi_application, **call):
        assert response.status == '200 OK'

        modified_call = when(
            'Trying different url!',
            url='/apiv1/devices?a=b&c=d'
        )
        assert modified_call.url_parameters is None
        assert response.status == 200
        assert response.json['query'] == 'a=b&c=d'


def test_authorization():
    def wsgi_application(environ, start_response):
        result = {}
        identity = environ.get('HTTP_AUTHORIZATION')
        if identity:
            result['identity'] = identity
        start_response('200 OK', [
            ('Content-Type', 'application/json;charset=utf-8'),
        ])

        yield json.dumps(result).encode()


    with Given(
        wsgi_application, title='Testing authorization header',
        url='/',
        authorization='testuser'
    ):
        assert response.json['identity'] == 'testuser'




import threading


thread_local = threading.local()


class ContextIsNotInitializedError(Exception):
    pass


class ContextStack(list):

    def push(self, item):
        self.append(item)

    @property
    def hasitem(self):
        return self

    @property
    def current(self):
        if not self.hasitem:
            raise ContextIsNotInitializedError()
        return self[-1]


class Context:
    thread_local_key = 'bddrest_context_stack'

    @classmethod
    def __context_stack__(cls):
        if not hasattr(thread_local, cls.thread_local_key):
            setattr(thread_local, cls.thread_local_key, ContextStack())
        return getattr(thread_local, cls.thread_local_key)

    def __enter__(self):
        self.__context_stack__().push(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        stack = self.__context_stack__()
        if stack.hasitem:
            stack.pop()

    @classmethod
    def get_current(cls):
        return cls.__context_stack__().current





class ObjectProxy:
    """
    A simple object proxy to let deferred object's initialize later (for example: just after import):
    This class encapsulates some tricky codes to resolve the proxied object members using the
    `__getattribute__` and '__getattr__'. SO TAKE CARE about modifying the code, to prevent
    infinite loops and stack-overflow situations.

    Module: fancy_module

        deferred_object = None  # Will be initialized later.
        def init():
            global deferred_object
            deferred_object = AnyValue()
        proxy = ObjectProxy(lambda: deferred_object)

    In another module:

        from fancy_module import proxy, init
        def my_very_own_function():
            try:
                return proxy.any_attr_or_method()
            except: ObjectNotInitializedError:
                init()
                return my_very_own_function()

    """

    def __init__(self, resolver):
        object.__setattr__(self, '_resolver', resolver)

    @property
    def proxied_object(self):
        o = object.__getattribute__(self, '_resolver')()
        # if still is none, raise the exception
        if o is None:
            raise ValueError('Backend object is not initialized yet.')
        return o

    def __getattr__(self, key):
        return getattr(object.__getattribute__(self, 'proxied_object'), key)

    def __setattr__(self, key, value):
        setattr(object.__getattribute__(self, 'proxied_object'), key, value)

    def __eq__(self, other):
        return self.proxied_object.__eq__(other)

    def __gt__(self, other):
        return self.proxied_object.__gt__(other)

    def __ge__(self, other):
        return self.proxied_object.__ge__(other)

    def __lt__(self, other):
        return self.proxied_object.__lt__(other)

    def __le__(self, other):
        return self.proxied_object.__le__(other)

    def __str__(self):
        return self.proxied_object.__str__()

    def __repr__(self):
        return self.proxied_object.__repr__()




import sys
import argparse
from os import path

import argcomplete


from .launchers import Launcher, RequireSubCommand


class MainLauncher(Launcher):

    def __init__(self):
        self.parser = parser = argparse.ArgumentParser(
            prog=path.basename(sys.argv[0]),
            description='bddrest command line interface.'
        )

        subparsers = parser.add_subparsers(
            title="Sub Commands",
            dest="command"
        )

        from ..documentary import DocumentaryLauncher
        DocumentaryLauncher.register(subparsers)

        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        cli_args = self.parser.parse_args(args if args else None)
        if hasattr(cli_args, 'func'):
            cli_args.func(cli_args)
        else:
            super().launch()


root_launcher = None


def main():
    global root_launcher
    if root_launcher is None:
        root_launcher = MainLauncher()
    return root_launcher()




from .call import Call
from .response import Response
from ..helpers import normalize_query_string
from .headerset import HeaderSet


class FirstCall(Call):

    _headers = None
    _url = None
    _url_parameters = None
    _verb = None
    _query = None
    _form = None
    _as = None
    _extra_environ = None

    # FIXME: remove them  and use header set to store it.
    _content_type = None
    _authorization = None


    def __init__(self, title: str, url='/', verb='GET',
                 url_parameters: dict = None, form: dict = None,
                 content_type: str = None, headers: list = None,
                 as_: str = None, query: dict = None, description: str = None,
                 extra_environ: dict = None, response: Response=None,
                 authorization: str = None):

        super().__init__(title, description=description, response=response)

        self.url = url
        # the `url_parameters` and `query` attributes may be set by the url setter. so we're
        # not going to override them anyway.
        if url_parameters is not None:
            self.url_parameters = url_parameters

        if query is not None:
            self.query = query

        self.verb = verb
        self.form = form
        self.content_type = content_type
        self.authorization = authorization
        self.headers = headers
        self.as_ = as_
        self.extra_environ = extra_environ

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url, self.url_parameters, self.query = self.extract_url_parameters(value)

    @property
    def url_parameters(self):
        return self._url_parameters

    @url_parameters.setter
    def url_parameters(self, value):
        self._url_parameters = value

    @property
    def verb(self):
        return self._verb

    @verb.setter
    def verb(self, value):
        self._verb = value

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = HeaderSet(value) if value is not None else None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, value):
        self._query = normalize_query_string(value)

    @property
    def content_type(self):
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        self._content_type = value

    @property
    def authorization(self):
        return self._authorization

    @authorization.setter
    def authorization(self, value):
        self._authorization = value

    @property
    def as_(self):
        return self._as

    @as_.setter
    def as_(self, value):
        self._as = value

    @property
    def extra_environ(self):
        return self._extra_environ

    @extra_environ.setter
    def extra_environ(self, value):
        self._extra_environ = value

    @property
    def form(self):
        return self._form

    @form.setter
    def form(self, value):
        self._form = value





import pytest

from bddrest.specification import HTTPStatus


def test_status_comparison():
    s = HTTPStatus('200 OK')
    assert s == '200 ok'
    assert s == 200
    assert s != 201
    assert s != '200 OKOK'
    assert s >= 100
    assert s > 100
    assert s <= 300
    assert s < 300
    assert s >= 200
    assert s <= 200

    with pytest.raises(ValueError):
        s >= '100 Continue'

    with pytest.raises(ValueError):
        s < '300 Multiple Choice'

    with pytest.raises(ValueError):
        s <= '300 Multiple Choice'

    with pytest.raises(ValueError):
        s > '100 Continue'



from .base import Formatter


class MarkdownFormatter(Formatter):
    def writeline(self, text=''):
        self.write(f'{text}\n')

    def write_header(self, text, level=1):
        self.writeline(f'{"#" * level} {text}\n')

    def write_paragraph(self, text):
        self.writeline(text)
        self.writeline()

    def _write_table_row(self, row):
        self.writeline(' | '.join(str(i) for i in row))

    def write_table(self, array2d, headers=None):
        if not isinstance(array2d, list):
            array2d = list(array2d)

        columns = len(array2d[0])
        if headers:
            self._write_table_row(headers)
        self.writeline(' | '.join(['---'] * columns))
        for row in array2d:
            self._write_table_row(row)

        self.writeline()

    def write_list(self, listkind):
        for l in listkind:
            self.writeline(f'* {l}')
        self.writeline()




from ..helpers import normalize_query_string
from .call import Call
from .headerset import HeaderSet
from .response import Response


class Unchanged:
    pass


UNCHANGED = Unchanged()


class AlteredCall(Call):
    def __init__(self, base_call, title, url=UNCHANGED, verb=UNCHANGED,
                 url_parameters=UNCHANGED, form=UNCHANGED,
                 content_type=UNCHANGED, headers=UNCHANGED, as_=UNCHANGED,
                 query=UNCHANGED, description=None, extra_environ=UNCHANGED,
                 response: Response=None, authorization=UNCHANGED):

        self.base_call = base_call
        self.diff = {}
        super().__init__(title, description=description, response=response)

        self.url = url
        if url_parameters is not UNCHANGED:
            self.url_parameters = url_parameters

        if query is not UNCHANGED:
            self.query = query

        self.extra_environ = extra_environ
        self.verb = verb
        self.form = form
        self.content_type = content_type
        self.authorization = authorization
        self.headers = headers
        self.as_ = as_

    def to_dict(self):
        result = dict(title=self.title)
        result.update(self.diff)

        if self.description is not None:
            result['description'] = self.description

        if self.response is not None:
            result['response'] = self.response.to_dict()

        return result

    def update_diff(self, key, value):
        if value is UNCHANGED:
            self.diff.pop(key, None)
            return

        self.diff[key] = value

    @property
    def url(self):
        return self.diff.get('url', self.base_call.url)

    @url.setter
    def url(self, value):
        if value is UNCHANGED:
            self.diff.pop('url', None)
            return

        url, url_parameters, query = self.extract_url_parameters(value)
        if url and url != self.base_call.url:
            self.diff['url'] = url
            self.url_parameters = url_parameters
            self.query = query

    @property
    def url_parameters(self):
        return self.diff.get('url_parameters', self.base_call.url_parameters)

    @url_parameters.setter
    def url_parameters(self, value):
        self.update_diff('url_parameters', value)

    @url_parameters.deleter
    def url_parameters(self):
        del self.diff['url_parameters']

    @property
    def verb(self):
        return self.diff.get('verb', self.base_call.verb)

    @verb.setter
    def verb(self, value):
        self.update_diff('verb', value)

    @verb.deleter
    def verb(self):
        del self.diff['verb']

    @property
    def headers(self):
        return self.diff.get('headers', self.base_call.headers)

    @headers.setter
    def headers(self, value):
        self.update_diff('headers', value if value is UNCHANGED else HeaderSet(value))

    @headers.deleter
    def headers(self):
        del self.diff['headers']

    @property
    def query(self):
        return self.diff.get('query', self.base_call.query)

    @query.setter
    def query(self, value):
        self.update_diff('query', value if value is UNCHANGED else normalize_query_string(value))

    @query.deleter
    def query(self):
        del self.diff['query']

    @property
    def content_type(self):
        return self.diff.get('content_type', self.base_call.content_type)

    @content_type.setter
    def content_type(self, value):
        self.update_diff('content_type', value)

    @content_type.deleter
    def content_type(self):
        del self.diff['content_type']

    @property
    def authorization(self):
        return self.diff.get('authorization', self.base_call.authorization)

    @authorization.setter
    def authorization(self, value):
        self.update_diff('authorization', value)

    @authorization.deleter
    def authorization(self):
        del self.diff['authorization']

    @property
    def as_(self):
        return self.diff.get('as_', self.base_call.as_)

    @as_.setter
    def as_(self, value):
        self.update_diff('as_', value)

    @as_.deleter
    def as_(self):
        del self.diff['as_']

    @property
    def extra_environ(self):
        return self.diff.get('extra_environ', self.base_call.extra_environ)

    @extra_environ.setter
    def extra_environ(self, value):
        self.update_diff('extra_environ', value)

    @extra_environ.deleter
    def extra_environ(self):
        del self.diff['extra_environ']

    @property
    def form(self):
        return self.diff.get('form', self.base_call.form)

    @form.setter
    def form(self, value):
        self.update_diff('form', value)

    @form.deleter
    def form(self):
        del self.diff['form']




from abc import ABCMeta, abstractmethod


class Formatter(metaclass=ABCMeta):
    def __init__(self, outfile):
        self.file = outfile

    def write(self, text):
        self.file.write(text)

    @abstractmethod
    def writeline(self, text=''):  # pragma: no cover
        pass

    @abstractmethod
    def write_header(self, text, level=1):  # pragma: no cover
        pass

    @abstractmethod
    def write_paragraph(self, text):  # pragma: no cover
        pass

    @abstractmethod
    def write_table(self, array2d, headers=None):  # pragma: no cover
        pass

    @abstractmethod
    def write_list(self, listkind):  # pragma: no cover
        pass




import cgi
import json
import pytest

from bddrest import Given, Append, Remove, Update, when, response, status


def wsgi_application(environ, start_response):
    result = {}
    start_response('200 OK', [
        ('Content-Type', 'application/json;charset=utf-8'),
    ])

    for key in  [h for h in environ if h.startswith('HTTP')]:
        result[key[5:].lower()] = environ[key]

    yield json.dumps(result).encode()


def test_append_headers_field():
    with Given(
        wsgi_application,
        'test add header field',
        headers={'header1': '1'}
    ):
        assert status == '200 OK'
        assert response.json['header1'] == '1'

        # Using dictionary to manipulate lists, expecting error.
        with pytest.raises(ValueError):
            when('Abusing', headers=Append(header2='2'))

        when(
            'Adding new header: 2-tuple',
            headers=Append(('header2', '2'))
        )
        assert response.json['header1'] == '1'
        assert response.json['header2'] == '2'

        when(
            'Adding new header: single string',
            headers=Append('header3: 3')
        )
        assert response.json['header1'] == '1'
        assert response.json['header3'] == '3'
        assert 'header2' not in response.json


def test_remove_headers_field():
    with Given(
        wsgi_application,
        'test remove  header field',
        headers={'header1': '1'}
    ):
        assert status == '200 OK'
        assert response.json['header1'] == '1'

        when(
            'Removing an existing header: 2-tuple',
            headers=Remove(('header1', '1'))
        )
        assert 'header1' not in response.json


        when('Remove header by key', headers=Remove('header1'))

        # Remove an invalid header(Not exists)
        with pytest.raises(ValueError):
            when('Invalid  key', headers=Remove('invalid header'))




import tempfile
import unittest
from os import path
import io

from bddrest import Given, when, response


def wsgi_application(environ, start_response):
    start_response('200 OK', [])
    yield b'Nothing'


def test_autodoc_filename():
    filename = tempfile.mktemp()
    with Given(
        wsgi_application,
        title='Testing auto documentation',
        url='/apiv1/devices/name: SM-12345678',
        autodoc=filename,
    ):
        assert response.status == 200

    assert path.exists(filename)


def test_autodoc_file_object():
    file = io.StringIO()
    with Given(
        wsgi_application,
        title='Testing auto documentation',
        url='/apiv1/devices/name: SM-12345678',
        autodoc=file,
    ):
        assert response.status == 200

    assert len(file.getvalue()) > 0





class Documenter:
    def __init__(self, formatter_factory):
        self.formatter_factory = formatter_factory

    def write_response(self, formatter, response):
        formatter.write_header(f'Response: {response.status}', 3)

        if response.headers:
            formatter.write_header('Headers', 4)
            formatter.write_list(f'{k}: {v}' for k, v in response.headers)

        if response.body:
            formatter.write_header('Body', 4)
            mime = ''
            if response.content_type and 'json' in response.content_type:
                mime = 'json'
            formatter.write_paragraph(f'```{mime}\n{response.text}\n```')

    def document(self, story, outfile):
        basecall = story.base_call
        formatter = self.formatter_factory(outfile)
        formatter.write_header(basecall.title, 2)
        formatter.write_header(f'{basecall.verb} {basecall.url}', 3)
        if basecall.description:
            formatter.write_paragraph(basecall.description)

        if basecall.url_parameters:
            formatter.write_header('Url Parameters', 3)
            formatter.write_table(basecall.url_parameters.items(), headers=('Name', 'Example'))

        if basecall.query:
            formatter.write_header('Query Strings', 3)
            formatter.write_table(basecall.query.items(), headers=('Name', 'Example'))

        if basecall.form:
            formatter.write_header('Form', 3)
            formatter.write_table(basecall.form.items(), headers=('Name', 'Example'))

        if basecall.headers:
            formatter.write_header('Request Headers', 3)
            formatter.write_list(f'{k}: {v}' for k, v in basecall.headers)

        if basecall.response:
            self.write_response(formatter, basecall.response)

        for call in story.calls:
            formatter.write_header(f'WHEN: {call.title}', 2)

            if call.description:
                formatter.write_paragraph(call.description)

            if call.url_parameters and call.url_parameters != basecall.url_parameters:
                formatter.write_header('Url Parameters', 3)
                formatter.write_table(call.url_parameters.items(), headers=('Name', 'Example'))

            if call.query and call.query != basecall.query:
                formatter.write_header('Query Strings', 3)
                formatter.write_table(call.query.items(), headers=('Name', 'Example'))

            if call.form and call.form != basecall.form:
                formatter.write_header('Form', 3)
                formatter.write_table(call.form.items(), headers=('Name', 'Example'))

            if call.headers and call.headers != basecall.headers:
                formatter.write_header('Request Headers', 3)
                formatter.write_list(f'{k}: {v}' for k, v in call.headers)

            if call.response:
                self.write_response(formatter, call.response)




import re
from os.path import join, dirname
from setuptools import setup, find_packages


# reading package version (without loading it)
with open(join(dirname(__file__), 'bddrest', '__init__.py')) as v_file:
    package_version = re.compile(r".*__version__ = '(.*?)'", re.S)\
        .match(v_file.read()).group(1)


dependencies = [
    'webtest',
    'pyyaml',
    'argcomplete'
]


setup(
    name='bddrest',
    version=package_version,
    description='A toolchain for testing REST APIs in BDD manner.',
    long_description=open('README.md').read(),
    author='Vahid Mardani',
    author_email='vahid.mardani@gmail.com',
    install_requires=dependencies,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'bddrest = bddrest.cli:main'
        ]
    },
    test_suite='bddrest.tests',
    license='MIT',
    url='https://github.com/Carrene/bddrest',
    classifiers=[
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]

)



import abc
from collections import Iterable

from .given import Given


class Manipulator(metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwargs):
        self.list_diff = args

        if not isinstance(self.list_diff, list):
            self.list_diff = list(self.list_diff)

        self.dict_diff = kwargs

    @abc.abstractmethod
    def apply(self):
        pass

    def __add__(self, other):
        return CompositeManipulator(self) + other

    def __sub__(self, other):
        return CompositeManipulator(self) - other

    def __or__(self, other):
        return CompositeManipulator(self) | other



class Append(Manipulator):
    def apply(self, container):
        if isinstance(container, dict):
            for k, v in self.dict_diff.items():
                if k in container:
                    raise ValueError(
                        f'The key: {k} is already exists in the target '
                        f'container. You may use the bddrest.Update object.'
                    )
            container.update(self.dict_diff)
            return

        if not self.list_diff:
            raise ValueError(
                'Invalid list manipulation value, Use positional argument '
                'instead of keyword arguments'
            )
        container.extend(self.list_diff)


class Update(Manipulator):
    def apply(self, container):
        if not isinstance(container, dict):
            raise ValueError('Only dict is supported for Update manipulator')

        container.update(self.dict_diff)


class Remove(Manipulator):
    def apply(self, container):
        if not isinstance(self.list_diff, list):
            raise ValueError('Only list is supported for Remove manipulator')

        for k in self.list_diff:
            if k not in container:
                raise ValueError(f'The key: {k} is not exist in the target')

            if isinstance(container, dict):
                del container[k]
            else:
                container.remove(k)


class CompositeManipulator(Manipulator):
    def __init__(self, *args):
        super().__init__()
        self.rules = list(args)

    def apply(self, container):
        for rule in self.rules:
            rule.apply(container)

    def __add__(self, other):
        if isinstance(other, dict):
            manipulator = Append(**other)
        elif isinstance(other, Manipulator):
            manipulator = other
        else:
            raise TypeError('Only dict or Manipulator will be accepted')

        self.rules.append(manipulator)
        return self

    def __sub__(self, other):
        if isinstance(other, str):
            manipulator = Remove(other)
        if isinstance(other, Iterable):
            manipulator = Remove(*other)
        elif isinstance(other, Manipulator):
            manipulator = other
        else:
            raise TypeError('Only str or an iterable of str will be accepted')

        self.rules.append(manipulator)
        return self

    def __or__(self, other):
        if isinstance(other, dict):
            manipulator = Update(**other)
        elif isinstance(other, Manipulator):
            manipulator = other
        else:
            raise TypeError('Only dict or Manipulator will be accepted')

        self.rules.append(manipulator)
        return self


class CompositeManipulatorInitializer(CompositeManipulator):
    def __add__(self, other):
        return CompositeManipulator() + other

    def __sub__(self, other):
        return CompositeManipulator() - other

    def __or__(self, other):
        return CompositeManipulator() | other


given_form = CompositeManipulatorInitializer()


def when(*args, **kwargs):
    story = Given.get_current()

    # Checking for list manipulators if any
    # Checking for dictionary manipulators if any
    for k, v in kwargs.items():
        if isinstance(v, Manipulator):
            clone = getattr(story.base_call, k).copy()
            v.apply(clone)
            kwargs[k] = clone

    return story.when(*args, **kwargs)




import cgi
import json

from bddrest import Given, when, response


def wsgi_application(environ, start_response):
    form = cgi.FieldStorage(
        fp=environ['wsgi.input'],
        environ=environ,
        strict_parsing=False,
        keep_blank_values=True
    )

    try:
        code = int(form['activationCode'].value) ^ 1234
    except ValueError:
        start_response('400 Bad Request', [('Content-Type', 'text/plain;utf-8')])
        return

    start_response('200 OK', [
        ('Content-Type', 'application/json;charset=utf-8'),
        ('X-Pagination-Count', '10')
    ])
    result = json.dumps(dict(
        secret='ABCDEF',
        code=code,
        query=environ['QUERY_STRING']
    ))
    yield result.encode()


def test_equality():

    call = dict(
        title=\
            'Binding and registering the device after verifying the '
            'activation code',
        description=\
            'As a new visitor I have to bind my device with activation '
            'code and phone number',
        url='/apiv1/devices/name: SM-12345678',
        verb='POST',
        as_='visitor',
        query=dict(
            a=1,
            b=2
        ),
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        )
    )
    with Given(wsgi_application, **call):
        assert response.status == '200 OK'
        assert response.status == 200

        when(
            'Trying invalid code',
            form=dict(
                activationCode='badCode'
            )
        )
        assert response.status == 400

